/*
 * Copyright (C) 1997 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file server.h Common server data structures. */

/** Size of the selector_info and audio_file info strings of struct misc_meta_data. */
#define MMD_INFO_SIZE 16384

/** The maximum length of the host component in an URL */
#define MAX_HOSTLEN 256


/** Arguments for the sender command. */
struct sender_command_data {
	/** Greater than zero indicates that a sender cmd is already queued. */
	int cmd_num;
	/** The number of the sender in question. */
	int sender_num;
	/** Used for the allow/deny/add/remove subcommands. */
	char host[MAX_HOSTLEN];
	/** Used for allow/deny. */
	int netmask;
	/** The port number for add/remove. */
	int port;
	/** Maximal slice size. */
	uint16_t max_slice_bytes;
	/** Number of data slices plus redundant slices. */
	uint8_t slices_per_group;
	/** Number of slices minus number of redundant slices. */
	uint8_t data_slices_per_group;
};

/**
 * Miscellaneous data for communication between server and command handlers.
 *
 * There's only one instance of this structure which lives in a shared memory
 * area. Command handlers communicate with the server process through this
 * area. Changes made by the command handlers stay after the command handler
 * exits. Conversely, changes made by the server process propagate to the
 * command handlers. Access to this area is serialized via mmd_lock() and
 * mmd_unlock().
 *
 * There are two reasons for a variable to be included here: (a) at least one
 * command handler changes its value, or (b) updates by the server must
 * propagate to the stat command handlers.
 */
struct misc_meta_data {
	/** The size of the current audio file in bytes. */
	size_t size;
	/** The "old" status flags -- commands may only read them. */
	unsigned int vss_status_flags;
	/** The new status flags -- commands may set them. */
	unsigned int new_vss_status_flags;
	/** The number of data chunks sent so far. */
	long unsigned chunks_sent;
	/** Set by the jmp/ff commands to the new position in chunks. */
	long unsigned repos_request;
	/** The number of the chunk currently being sent. */
	long unsigned current_chunk;
	/** The milliseconds that have been skipped of the current audio file. */
	long offset;
	/** The time para_server started to stream. */
	struct timeval stream_start;
	/**
	 * The event counter.
	 *
	 * Commands may increase this to force a status update to be sent to
	 * all connected stat clients.
	 */
	unsigned int events;
	/** The number of audio files already sent. */
	unsigned int num_played;
	/** The number of executed commands. */
	unsigned int num_commands;
	/** The number of connections para_server received so far. */
	unsigned int num_connects;
	/** The number of connections currently active. */
	unsigned int active_connections;
	/** The process id of the audio file selector. */
	pid_t afs_pid;
	/** This gets updated by afs and contains its current mode. */
	char afs_mode_string[MAXLINE];
	/** Used by the sender command. */
	struct sender_command_data sender_cmd_data;
	/** Describes the current audio file. */
	struct audio_file_data afd;
};

/** Command line options for para_server. */
extern struct server_args_info conf;

__noreturn void handle_connect(int fd, const char *peername);
void parse_config_or_die(int override);
char *server_get_tasks(void);
